import os

import pytest

from cube_wrangler.roadway import Parameters, ModelRoadwayNetwork
from network_wrangler import RoadwayNetwork

STPAUL_DIR = os.path.join(os.getcwd(), "data")

STPAUL_SHAPE_FILE = os.path.join(STPAUL_DIR, "st_paul_shape.geojson")
STPAUL_LINK_FILE = os.path.join(STPAUL_DIR, "st_paul_link.json")
STPAUL_NODE_FILE = os.path.join(STPAUL_DIR, "st_paul_node.geojson")


@pytest.mark.roadway
@pytest.mark.travis
def test_parameter_read(request):
    """
    Tests that parameters are read
    """
    print("\n--Starting:", request.node.name)

    params = Parameters()
    print(params.__dict__)
    ## todo write an assert that actually tests something


@pytest.mark.roadway
@pytest.mark.travis
def test_network_calculate_variables(request):
    """
    Tests that parameters are read
    """
    print("\n--Starting:", request.node.name)

    net = ModelRoadwayNetwork.read(
        link_file=STPAUL_LINK_FILE,
        node_file=STPAUL_NODE_FILE,
        shape_file=STPAUL_SHAPE_FILE,
        fast=True,
    )

    assert isinstance(net, ModelRoadwayNetwork)
    assert len(net.links_df) == 58821



"""Transit-related classes to parse, compare, and write standard and cube transit files.

  Typical usage example:

    tn = CubeTransit.create_from_cube(CUBE_DIR)
    transit_change_list = tn.evaluate_differences(base_transit_network)

    cube_transit_net = StandardTransit.read_gtfs(BASE_TRANSIT_DIR)
    cube_transit_net.write_as_cube_lin(os.path.join(WRITE_DIR, "outfile.lin"))
"""
import os
import copy
import csv
import datetime, time
from typing import Any, Dict, Optional, Union

from lark import Lark, Transformer, v_args
from pandas import DataFrame

import pandas as pd
import partridge as ptg

from network_wrangler.transit.network import TransitNetwork

from .logger import WranglerLogger
from .parameters import Parameters


class StandardTransit(object):
    """Holds a standard transit feed as a Partridge object and contains
    methods to manipulate and translate the GTFS data to MetCouncil's
    Cube Line files.

    .. highlight:: python
    Typical usage example:
    ::
        cube_transit_net = StandardTransit.read_gtfs(BASE_TRANSIT_DIR)
        cube_transit_net.write_as_cube_lin(os.path.join(WRITE_DIR, "outfile.lin"))

    Attributes:
        feed: Partridge Feed object containing read-only access to GTFS feed
        parameters (Parameters): Parameters instance containing information
            about time periods and variables.
    """

    def __init__(
        self, ptg_feed, road_net=None, parameters: Union[Parameters, dict] = {}
    ):
        """

        Args:
            ptg_feed: partridge feed object
            parameters: dictionary of parameter settings (see Parameters class) or an instance of Parameters
        """
        self.feed = ptg_feed
        self.road_net = road_net

        if type(parameters) is dict:
            self.parameters = Parameters(**parameters)
        elif isinstance(parameters, Parameters):
            # self.parameters = Parameters(**parameters.__dict__)
            self.parameters = parameters
        else:
            msg = "Parameters should be a dict or instance of Parameters: found {} which is of type:{}".format(
                parameters, type(parameters)
            )
            WranglerLogger.error(msg)
            raise ValueError(msg)

    @staticmethod
    def fromTransitNetwork(
        transit_network_object: TransitNetwork, parameters: Union[Parameters, dict] = {}
    ):
        """
        RoadwayNetwork to ModelRoadwayNetwork

        Args:
            transit_network_object: Reference to an instance of TransitNetwork.
            parameters: dictionary of parameter settings (see Parameters class) or an instance of Parameters. If not provided will
                use default parameters.

        Returns:
            StandardTransit
        """
        return StandardTransit(
            transit_network_object.feed,
            transit_network_object.road_net,
            parameters=parameters,
        )

    @staticmethod
    def read_gtfs(gtfs_feed_dir: str, parameters: Union[Parameters, dict] = {}):
        """
        Reads GTFS files from a directory and returns a StandardTransit
        instance.

        Args:
            gtfs_feed_dir: location of the GTFS files
            parameters: dictionary of parameter settings (see Parameters class) or an instance of Parameters. If not provided will
                use default parameters.

        Returns:
            StandardTransit instance
        """
        return StandardTransit(ptg.load_feed(gtfs_feed_dir), parameters=parameters)

    def write_as_cube_lin(self, outpath: str = None, line_name_xwalk: str = None):
        """
        Writes the gtfs feed as a cube line file after
        converting gtfs properties to MetCouncil cube properties.

        Args:
            outpath: File location for output cube line file.

        """
        if not outpath:
            outpath = os.path.join(self.parameters.scratch_location, "outtransit.lin")
        # trip_cube_df = self.route_properties_gtfs_to_cube(self, line_name_xwalk)

        # trip_cube_df["LIN"] = trip_cube_df.apply(self.cube_format, axis=1)

        l = self.feed.trip_cube_df["LIN"].tolist()
        l = [";;<<PT>><<LINE>>;;"] + l

        with open(outpath, "w") as f:
            f.write("\n".join(l))

    def time_to_cube_time_period(
        self, start_time_secs: int, as_str: bool = True, verbose: bool = False
    ):
        """
        Converts seconds from midnight to the cube time period.

        Args:
            start_time_secs: start time for transit trip in seconds
                from midnight
            as_str: if True, returns the time period as a string,
                otherwise returns a numeric time period

        Returns:
            this_tp_num: if as_str is False, returns the numeric
                time period
            this_tp: if as_str is True, returns the Cube time period
                name abbreviation
        """
        from .util import hhmmss_to_datetime, secs_to_datetime

        # set initial time as the time that spans midnight

        start_time_dt = secs_to_datetime(start_time_secs)

        # set initial time as the time that spans midnight
        this_tp = "NA"
        for tp_name, _times in self.parameters.time_period_to_time.items():
            _start_time, _end_time = _times
            _dt_start_time = hhmmss_to_datetime(_start_time)
            _dt_end_time = hhmmss_to_datetime(_end_time)
            if _dt_start_time > _dt_end_time:
                this_tp = tp_name
                break

        for tp_name, _times in self.parameters.time_period_to_time.items():
            _start_time, _end_time = _times
            _dt_start_time = hhmmss_to_datetime(_start_time)
            if start_time_dt >= _dt_start_time:
                this_time = _dt_start_time
                this_tp = tp_name

        if verbose:
            WranglerLogger.debug(
                "Finding Cube Time Period from Start Time: \
                \n  - start_time_sec: {} \
                \n  - start_time_dt: {} \
                \n  - this_tp: {}".format(
                    start_time_secs, start_time_dt, this_tp
                )
            )

        if as_str:
            return this_tp

        name_to_num = {v: k for k, v in self.parameters.cube_time_periods.items}
        this_tp_num = name_to_num.get(this_tp)

        if not this_tp_num:
            msg = (
                "Cannot find time period number in {} for time period name: {}".format(
                    name_to_num, this_tp
                )
            )
            WranglerLogger.error(msg)
            raise ValueError(msg)

        return this_tp_num


class CubeTransformer(Transformer):
    """A lark-parsing Transformer which transforms the parse-tree to
    a dictionary.

    .. highlight:: python
    Typical usage example:
    ::
        transformed_tree_data = CubeTransformer().transform(parse_tree)

    Attributes:
        line_order (int): a dynamic counter to hold the order of the nodes within
            a route shape
        lines_list (list): a list of the line names
    """

    def __init__(self):
        self.line_order = 0
        self.lines_list = []

    def lines(self, line):
        # WranglerLogger.debug("lines: \n {}".format(line))

        # This MUST be a tuple because it returns to start in the tree
        lines = {k: v for k, v in line}
        return ("lines", lines)

    @v_args(inline=True)
    def program_type_line(self, PROGRAM_TYPE, whitespace=None):
        # WranglerLogger.debug("program_type_line:{}".format(PROGRAM_TYPE))
        self.program_type = PROGRAM_TYPE.value

        # This MUST be a tuple because it returns to start  in the tree
        return ("program_type", PROGRAM_TYPE.value)

    @v_args(inline=True)
    def line(self, lin_attributes, nodes):
        # WranglerLogger.debug("line...attributes:\n  {}".format(lin_attributes))
        # WranglerLogger.debug("line...nodes:\n  {}".format(nodes))
        lin_name = lin_attributes["NAME"]

        self.line_order = 0
        # WranglerLogger.debug("parsing: {}".format(lin_name))

        return (lin_name, {"line_properties": lin_attributes, "line_shape": nodes})

    @v_args(inline=True)
    def lin_attributes(self, *lin_attr):
        lin_attr = {k: v for (k, v) in lin_attr}
        # WranglerLogger.debug("lin_attributes:  {}".format(lin_attr))
        return lin_attr

    @v_args(inline=True)
    def lin_attr(self, lin_attr_name, attr_value, SEMICOLON_COMMENT=None):
        # WranglerLogger.debug("lin_attr {}:  {}".format(lin_attr_name, attr_value))
        return lin_attr_name, attr_value

    def lin_attr_name(self, args):
        attr_name = args[0].value.upper()
        # WranglerLogger.debug(".......args {}".format(args))
        if attr_name in ["USERA", "FREQ", "HEADWAY"]:
            attr_name = attr_name + "[" + str(args[2]) + "]"
        return attr_name

    def attr_value(self, attr_value):
        try:
            return int(attr_value[0].value)
        except:
            return attr_value[0].value

    def nodes(self, lin_node):
        lin_node = DataFrame(lin_node)
        # WranglerLogger.debug("nodes:\n {}".format(lin_node))

        return lin_node

    @v_args(inline=True)
    def lin_node(self, NODE_NUM, SEMICOLON_COMMENT=None, *lin_nodeattr):
        self.line_order += 1
        n = int(NODE_NUM.value)
        return {"node_id": abs(n), "node": n, "stop": n > 0, "order": self.line_order}

    start = dict


TRANSIT_LINE_FILE_GRAMMAR = r"""

start             : program_type_line? lines
WHITESPACE        : /[ \t\r\n]/+
STRING            : /("(?!"").*?(?<!\\)(\\\\)*?"|'(?!'').*?(?<!\\)(\\\\)*?')/i
SEMICOLON_COMMENT : /;[^\n]*/
BOOLEAN           : "T"i | "F"i
program_type_line : ";;<<" PROGRAM_TYPE ">><<LINE>>;;" WHITESPACE?
PROGRAM_TYPE      : "PT" | "TRNBUILD"

lines             : line*
line              : "LINE" lin_attributes nodes

lin_attributes    : lin_attr+
lin_attr          : lin_attr_name "=" attr_value "," SEMICOLON_COMMENT*
TIME_PERIOD       : "1".."5"
!lin_attr_name     : "allstops"i
                    | "color"i
                    | ("freq"i "[" TIME_PERIOD "]")
                    | ("headway"i "[" TIME_PERIOD "]")
                    | "mode"i
                    | "name"i
                    | "oneway"i
                    | "owner"i
                    | "runtime"i
                    | "timefac"i
                    | "xyspeed"i
                    | "longname"i
                    | "shortname"i
                    | ("usera"i TIME_PERIOD)
                    | ("usern2"i)
                    | "circular"i
                    | "vehicletype"i
                    | "operator"i
                    | "faresystem"i

attr_value        : BOOLEAN | STRING | SIGNED_INT

nodes             : lin_node+
lin_node          : ("N" | "NODES")? "="? NODE_NUM ","? SEMICOLON_COMMENT? lin_nodeattr*
NODE_NUM          : SIGNED_INT
lin_nodeattr      : lin_nodeattr_name "=" attr_value ","? SEMICOLON_COMMENT*
!lin_nodeattr_name : "access_c"i
                    | "access"i
                    | "delay"i
                    | "xyspeed"i
                    | "timefac"i
                    | "nntime"i
                    | "time"i

operator          : SEMICOLON_COMMENT* "OPERATOR" opmode_attr* SEMICOLON_COMMENT*
mode              : SEMICOLON_COMMENT* "MODE" opmode_attr* SEMICOLON_COMMENT*
opmode_attr       : ( (opmode_attr_name "=" attr_value) ","?  )
opmode_attr_name  : "number" | "name" | "longname"

%import common.SIGNED_INT
%import common.WS
%ignore WS

"""



import os
from .logger import WranglerLogger


def get_base_dir(cube_wrangler_base_dir=os.getcwd()):
    d = cube_wrangler_base_dir
    for i in range(3):
        if "conda-environments" in os.listdir(d):
            WranglerLogger.info("cube_wrangler base directory set as: {}".format(d))
            return d
        d = os.path.dirname(d)

    msg = "Cannot find cube_wrangler base directory from {}, please input using keyword in parameters: `cube_wrangler_base_dir =` ".format(
        cube_wrangler_base_dir
    )
    WranglerLogger.error(msg)
    raise (ValueError(msg))


class Parameters:
    """A class representing all the parameters defining the networks
    including time of day, categories, etc.

    Parameters can be set at runtime by initializing a parameters instance
    with a keyword argument setting the attribute.  Parameters that are
    not explicitly set will use default parameters listed in this class.
    .. highlight:: python
    ##TODO potentially split this between several classes.

    Attr:
    """

    def __init__(self, **kwargs):
        """
        constructor for the Parameters class
        """
        if "cube_wrangler_base_dir" in kwargs:
            self.base_dir = get_base_dir(
                cube_wrangler_base_dir=kwargs.get("cube_wrangler_base_dir")
            )
        else:
            self.base_dir = get_base_dir()

        if "settings_location" in kwargs:
            self.settings_location = kwargs.get("settings_location")
        else:
            self.settings_location = os.path.join(self.base_dir, "examples", "settings")

        if "scratch_location" in kwargs:
            self.scratch_location = kwargs.get("scratch_location")
        else:
            self.scratch_location = os.path.join(self.base_dir, "tests", "scratch")

        if "time_periods_to_time" in kwargs:
            self.time_periods_to_time = kwargs.get("time_periods_to_time")
        else:
            self.time_period_to_time = {
                "EA": ("3:00", "6:00"),
                "AM": ("6:00", "10:00"),  ##TODO FILL IN with real numbers
                "MD": ("10:00", "15:00"),
                "PM": ("15:00", "19:00"),
                "NT": ("19:00", "3:00"),
            }

        if "categories" in kwargs:
            self.categories = kwargs.get("categories")
        else:
            self.categories = {
                # suffix, source (in order of search)
                "sov": ["sov", "default"],
                "hov2": ["hov2", "default", "sov"],
                "hov3": ["hov3", "hov2", "default", "sov"],
                "truck": ["trk", "sov", "default"],
            }

        # prefix, source variable, categories
        self.properties_to_split = {
            "trn_priority": {
                "v": "trn_priority",
                "time_periods": self.time_period_to_time,
            },
            "ttime_assert": {
                "v": "ttime_assert",
                "time_periods": self.time_period_to_time,
            },
            "lanes": {"v": "lanes", "time_periods": self.time_period_to_time},
            "ML_lanes": {"v": "ML_lanes", "time_periods": self.time_period_to_time},
            "price": {
                "v": "price",
                "time_periods": self.time_period_to_time,
                "categories": self.categories,
            },
            "access": {"v": "access", "time_periods": self.time_period_to_time},
        }

        self.net_to_dbf_crosswalk = os.path.join(
            self.settings_location, "net_to_dbf.csv"
        )

        self.log_to_net_crosswalk = os.path.join(
            self.settings_location, "log_to_net.csv"
        )

        self.output_variables = [
            "model_link_id",
            "link_id",
            "A",
            "B",
            "shstGeometryId",
            "shape_id",
            "distance",
            "roadway",
            "name",
            "roadway_class",
            "bike_access",
            "walk_access",
            "drive_access",
            "truck_access",
            "trn_priority_AM",
            "trn_priority_MD",
            "trn_priority_PM",
            "trn_priority_NT",
            "ttime_assert_AM",
            "ttime_assert_MD",
            "ttime_assert_PM",
            "ttime_assert_NT",
            "lanes_AM",
            "lanes_MD",
            "lanes_PM",
            "lanes_NT",
            "price_sov_AM",
            "price_hov2_AM",
            "price_hov3_AM",
            "price_truck_AM",
            "price_sov_MD",
            "price_hov2_MD",
            "price_hov3_MD",
            "price_truck_MD",
            "price_sov_PM",
            "price_hov2_PM",
            "price_hov3_PM",
            "price_truck_PM",
            "price_sov_NT",
            "price_hov2_NT",
            "price_hov3_NT",
            "price_truck_NT",
            "roadway_class_idx",
            "assign_group",
            "access_AM",
            "access_MD",
            "access_PM",
            "access_NT",
            "mpo",
            "area_type",
            "county",
            "centroidconnect",
            "AADT",
            "count_year",
            "count_AM",
            "count_MD",
            "count_PM",
            "count_NT",
            "count_daily",
            "model_node_id",
            "N",
            "osm_node_id",
            "bike_node",
            "transit_node",
            "walk_node",
            "drive_node",
            "geometry",
            "X",
            "Y",
            "ML_lanes_AM",
            "ML_lanes_MD",
            "ML_lanes_PM",
            "ML_lanes_NT",
            "segment_id",
            "managed",
            "bus_only",
            "rail_only",
            "bike_facility",
            "bike",
            "walk",
        ]        

        self.output_link_shp = os.path.join(self.scratch_location, "links.shp")
        self.output_node_shp = os.path.join(self.scratch_location, "nodes.shp")
        self.output_link_csv = os.path.join(self.scratch_location, "links.csv")
        self.output_node_csv = os.path.join(self.scratch_location, "nodes.csv")
        self.output_link_txt = os.path.join(self.scratch_location, "links.txt")
        self.output_node_txt = os.path.join(self.scratch_location, "nodes.txt")
        self.output_link_header_width_txt = os.path.join(
            self.scratch_location, "links_header_width.txt"
        )
        self.output_node_header_width_txt = os.path.join(
            self.scratch_location, "nodes_header_width.txt"
        )
        self.output_cube_network_script = os.path.join(
            self.scratch_location, "make_complete_network_from_fixed_width_file.s"
        )

        self.bool_col = [
            "rail_only",
            "bus_only",
            "drive_access",
            "bike_access",
            "walk_access",
            "truck_access",
        ]

        self.int_col = [
            "model_link_id",
            "model_node_id",
            "A",
            "B",
            "lanes_AM",
            "lanes_MD",
            "lanes_PM",
            "lanes_NT",
            "roadway_class",
            "assign_group",
            "county",
            "area_type",
            "trn_priority",
            "AADT",
            "count_AM",
            "count_MD",
            "count_PM",
            "count_NT",
            "count_daily",
            "centroidconnect",
            "bike_facility",
            "truck_access",
            "drive_node",
            "walk_node",
            "bike_node",
            "transit_node",
            "ML_lanes_AM",
            "ML_lanes_MD",
            "ML_lanes_PM",
            "ML_lanes_NT",
            "segment_id",
            "managed",
            "bike",
            "walk",
        ]

        self.float_col = ["distance", "ttime_assert", "price", "X", "Y"]

        self.zones = 3061

        self.__dict__.update(kwargs)



import re
from .logger import WranglerLogger
from unidecode import unidecode

import numpy as np


def get_shared_streets_intersection_hash(lat, long, osm_node_id=None):
    """
    Calculated per:
       https://github.com/sharedstreets/sharedstreets-js/blob/0e6d7de0aee2e9ae3b007d1e45284b06cc241d02/src/index.ts#L553-L565
    Expected in/out
      -93.0965985, 44.952112199999995 osm_node_id = 954734870
       69f13f881649cb21ee3b359730790bb9

    """
    import hashlib

    message = "Intersection {0:.5f} {0:.5f}".format(long, lat)
    if osm_node_id:
        message += " {}".format(osm_node_id)
    unhashed = message.encode("utf-8")
    hash = hashlib.md5(unhashed).hexdigest()
    return hash


def hhmmss_to_datetime(hhmmss_str: str):
    """
    Creates a datetime time object from a string of hh:mm:ss

    Args:
        hhmmss_str: string of hh:mm:ss
    Returns:
        dt: datetime.time object representing time
    """
    import datetime

    dt = datetime.time(*[int(i) for i in hhmmss_str.split(":")])

    return dt


def secs_to_datetime(secs: int):
    """
    Creates a datetime time object from a seconds from midnight

    Args:
        secs: seconds from midnight
    Returns:
        dt: datetime.time object representing time
    """
    import datetime

    dt = (datetime.datetime.min + datetime.timedelta(seconds=secs)).time()

    return dt


def column_name_to_parts(c, parameters=None):

    if not parameters:
        from .parameters import Parameters

        parameters = Parameters()

    if c[0:2] == "ML":
        managed = 1
    else:
        managed = 0

    time_period = None
    category = None

    if c.split("_")[0] not in parameters.properties_to_split.keys():
        if c.split("_")[-1] in parameters.time_period_to_time.keys():
            time_period = c.split("_")[-1]
            base_name = c.split(time_period)[-2][:-1]
            if base_name not in parameters.properties_to_split.keys():
                return c, None, None, managed
        else:
            return c, None, None, managed

    tps = parameters.time_period_to_time.keys()
    cats = parameters.categories.keys()

    if c.split("_")[-1] in tps:
        time_period = c.split("_")[-1]
        base_name = c.split(time_period)[-2][:-1]
        if c.split("_")[-2] in cats:
            category = c.split("_")[-2]
            base_name = c.split(category)[-2][:-1]
    elif c.split("_")[-1] in cats:
        category = c.split("_")[-1]
        base_name = c.split(category)[-2][:-1]
    else:
        msg = "Can't split property correctly: {}".format(c)
        WranglerLogger.error(msg)

    return base_name, time_period, category, managed


def shorten_name(name):
    if type(name) == str:
        name_list = name.split(",")
    elif type(name) in [float, np.int32, np.int64]:
        name_list = str(name)
    else:
        name_list = name
    name_list = [
        re.sub(r"\W+", " ", str(c)).replace("nan", "").strip(" ") for c in name_list
    ]

    name_list = list(set(name_list))
    # name_list.remove('')

    name_new = " ".join(name_list).strip(" ")

    # convert non english character to english
    name_new = unidecode(name_new)

    return name_new



import json
import os
import re
from typing import Any, Dict, Optional, Union, List
from csv import reader
import numpy as np

import pandas as pd
from pandas import DataFrame
import geopandas as gpd

# from network_wrangler import ProjectCard
from projectcard import ProjectCard, write_card
from network_wrangler.roadway.network import RoadwayNetwork
from network_wrangler import load_roadway

from .transit import StandardTransit
from .logger import WranglerLogger
from .parameters import Parameters
from .roadway import split_properties_by_time_period_and_category
from .util import column_name_to_parts


class Project(object):
    """A single or set of changes to the roadway or transit system.

    Compares a base and a build transit network or a base and build
    highway network and produces project cards.

    .. highlight:: python

    Typical usage example:
    ::
        test_project = Project.create_project(
            base_transit_source=os.path.join(CUBE_DIR, "transit.LIN"),
            build_transit_source=os.path.join(CUBE_DIR, "transit_route_shape_change"),
        )
        test_project.evaluate_changes()
        test_project.write_project_card(
            os.path.join(SCRATCH_DIR, "t_transit_shape_test.yml")
        )

    Attributes:
        DEFAULT_PROJECT_NAME: a class-level constant that defines what
            the project name will be if none is set.
        STATIC_VALUES: a class-level constant which defines values that
            are not evaluated when assessing changes.
        card_data (dict):  {"project": <project_name>, "changes": <list of change dicts>}
        roadway_changes (DataFrame):  pandas dataframe of CUBE roadway changes.
        transit_changes (CubeTransit):
        base_roadway_network (RoadwayNetwork):
        base_transit_network (CubeTransit):
        build_transit_network (CubeTransit):
        project_name (str): name of the project, set to DEFAULT_PROJECT_NAME if not provided
        parameters: an  instance of the Parameters class which sets a bunch of parameters
    """

    DEFAULT_PROJECT_NAME = "USER TO define"

    STATIC_VALUES = [
        "model_link_id",
        "area_type",
        "county",
        # "assign_group",
        "centroidconnect",
    ]
    CALCULATED_VALUES = [
        "area_type",
        "county",
        "assign_group",
        "centroidconnect",
    ]

    def __init__(
        self,
        roadway_changes: Optional[DataFrame] = None,
        transit_changes: Optional[Any] = None,
        base_roadway_network: Optional[RoadwayNetwork] = None,
        base_transit_network: Optional[Any] = None,
        build_transit_network: Optional[Any] = None,
        project_name: Optional[str] = "",
        evaluate: Optional[bool] = False,
        parameters: Union[dict, Parameters] = {},
        # transit_shape_crosswalk_dict: dict = None,
    ):
        """
        ProjectCard constructor.

        args:
            roadway_changes: dataframe of roadway changes read from a log file
            transit_changes:
            base_roadway_network: RoadwayNetwork instance for base case
            base_transit_network: CubeTransit instance for base transit network
            build_transit_network: CubeTransit instance for build transit network
            project_name: name of the project
            evaluate: defaults to false, but if true, will create card data
            parameters: dictionary of parameter settings (see Parameters class) or an instance of Parameters. If not specified, will use default parameters.
            base_transit_shape_crosswalk_dict: dictionary of model shape id to standard shape id for transit shape crosswalk, if any

        returns: instance of ProjectCard
        """
        self.card_data = Dict[str, Dict[str, Any]]

        self.roadway_changes = roadway_changes
        self.base_roadway_network = base_roadway_network
        self.base_transit_network = base_transit_network
        self.build_transit_network = build_transit_network
        self.transit_changes = transit_changes
        self.project_name = (
            project_name if project_name else Project.DEFAULT_PROJECT_NAME
        )

        if type(parameters) is dict:
            self.parameters = Parameters(**parameters)
        elif isinstance(parameters, Parameters):
            # self.parameters = Parameters(**parameters.__dict__)
            self.parameters = parameters
        else:
            msg = "Parameters should be a dict or instance of Parameters: found {} which is of type:{}".format(
                parameters, type(parameters)
            )
            WranglerLogger.error(msg)
            raise ValueError(msg)

        if base_roadway_network != None:
            self.determine_roadway_network_changes_compatibility(
                self.base_roadway_network, self.roadway_changes, self.parameters
            )

        # if transit_shape_crosswalk_dict is not None:
        #     self.transit_shape_crosswalk_dict = transit_shape_crosswalk_dict

        if evaluate:
            self.evaluate_changes()

    def write_project_card(self, filename: str = None):
        """
        Writes project cards.

        Args:
            filename (str): File path to output .yml

        Returns:
            None
        """
        write_card(ProjectCard(self.card_data), filename=filename)

    @staticmethod
    def create_project(
        roadway_log_file: Union[str, List[str], None] = None,
        roadway_shp_file: Optional[str] = None,
        roadway_csv_file: Optional[str] = None,
        base_roadway_dir: Optional[str] = None,
        base_transit_source: Optional[str] = None,
        build_transit_source: Optional[str] = None,
        roadway_changes: Optional[DataFrame] = None,
        transit_changes: Optional[Any] = None,
        base_roadway_network: Optional[RoadwayNetwork] = None,
        base_transit_network: Optional[Any] = None,
        build_transit_network: Optional[Any] = None,
        project_name: Optional[str] = None,
        recalculate_calculated_variables: Optional[bool] = False,
        recalculate_distance: Optional[bool] = False,
        parameters: Optional[dict] = {},
        # transit_shape_crosswalk_file: Optional[str] = None,
        # model_shape_id_column: Optional[str] = "model_shape_id",
    ):
        """
        Constructor for a Project instance.

        Args:
            roadway_log_file: File path to consuming logfile or a list of logfile paths.
            roadway_shp_file: File path to consuming shape file for roadway changes.
            roadway_csv_file: File path to consuming csv file for roadway changes.
            base_roadway_dir: Folder path to base roadway network.
            base_transit_source: Folder path to base transit network or cube line file string.
            base_transit_file: File path to base transit network.
            build_transit_source: Folder path to build transit network or cube line file string.
            build_transit_file: File path to build transit network.
            roadway_changes: pandas dataframe of CUBE roadway changes.
            transit_changes: build transit changes.
            base_roadway_network: Base roadway network object.
            base_transit_network: Base transit network object.
            build_transit_network: Build transit network object.
            project_name:  If not provided, will default to the roadway_log_file filename if provided (or the first filename if a list is provided)
            recalculate_calculated_variables: if reading in a base network, if this is true it will recalculate variables such as area type, etc. This only needs to be true if you are creating project cards that are changing the calculated variables.
            recalculate_distance:  recalculate the distance variable. This only needs to be true if you are creating project cards that change the distance.
            parameters: dictionary of parameters
            transit_shape_crosswalk_file: File path to transit shape crosswalk file.
        Returns:
            A Project instance.
        """

        # if transit_shape_crosswalk_file:
        #     WranglerLogger.info(
        #         "Reading transit shape crosswalk file: {}".format(
        #             transit_shape_crosswalk_file
        #         )
        #     )
        #     transit_shape_crosswalk_df = pd.read_csv(transit_shape_crosswalk_file)
        #     WranglerLogger.info(
        #         "Will convert model shape id {} to standard shape_id".format(
        #             model_shape_id_column
        #         )
        #     )

        #     assert "shape_id" in transit_shape_crosswalk_df.columns, "shape_id not found in transit shape crosswalk file"
        #     assert model_shape_id_column in transit_shape_crosswalk_df.columns, "model shape id {} not found in transit shape crosswalk file".format(model_shape_id_column)

        #     transit_shape_crosswalk_dict = dict(
        #         zip(
        #             transit_shape_crosswalk_df[model_shape_id_column].astype(str),
        #             transit_shape_crosswalk_df["shape_id"].astype(str)
        #         )
        #     )
        # else:
        #     transit_shape_crosswalk_dict = None

        if base_transit_source and base_transit_network:
            msg = "Method takes only one of 'base_transit_source' and 'base_transit_network' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        # if base_transit_source:
        #     base_transit_network = CubeTransit.create_from_cube(
        #         base_transit_source,
        #         parameters,
        #         transit_shape_crosswalk_dict,
        #     )
        #     WranglerLogger.debug(
        #         "Base network has {} lines".format(len(base_transit_network.lines))
        #     )
        #     if len(base_transit_network.lines) <= 10:
        #         WranglerLogger.debug(
        #             "Base network lines: {}".format(
        #                 "\n - ".join(base_transit_network.lines)
        #             )
        #         )
        elif base_transit_network:
            pass
        else:
            msg = "No base transit network."
            WranglerLogger.info(msg)
            base_transit_network = None

        if build_transit_source and transit_changes:
            msg = "Method takes only one of 'build_transit_source' and 'transit_changes' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        # if build_transit_source:
        #     WranglerLogger.debug("build")
        #     build_transit_network = CubeTransit.create_from_cube(
        #         build_transit_source,
        #         parameters,
        #         transit_shape_crosswalk_dict,
        #     )
        #     WranglerLogger.debug(
        #         "Build network has {} lines".format(len(build_transit_network.lines))
        #     )
        #     if len(build_transit_network.lines) <= 10:
        #         WranglerLogger.debug(
        #             "Build network lines: {}".format(
        #                 "\n - ".join(build_transit_network.lines)
        #             )
        #         )
        elif transit_changes:
            pass
        else:
            msg = "No transit changes given or processed."
            WranglerLogger.info(msg)
            transit_changes = None

        if roadway_log_file and roadway_changes:
            msg = "Method takes only one of 'roadway_log_file' and 'roadway_changes' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        if roadway_shp_file and roadway_changes:
            msg = "Method takes only one of 'roadway_shp_file' and 'roadway_changes' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        if roadway_csv_file and roadway_changes:
            msg = "Method takes only one of 'roadway_csv_file' and 'roadway_changes' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        if roadway_log_file and roadway_csv_file:
            msg = "Method takes only one of 'roadway_log_file' and 'roadway_csv_file' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        if roadway_shp_file and roadway_csv_file:
            msg = "Method takes only one of 'roadway_shp_file' and 'roadway_csv_file' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        if roadway_log_file and roadway_shp_file:
            msg = "Method takes only one of 'roadway_log_file' and 'roadway_shp_file' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        if roadway_log_file and not project_name:
            if type(roadway_log_file) == list:
                project_name = os.path.splitext(os.path.basename(roadway_log_file[0]))[
                    0
                ]
                WranglerLogger.info(
                    "No Project Name - Using name of first log file in list"
                )
            else:
                project_name = os.path.splitext(os.path.basename(roadway_log_file))[0]
                WranglerLogger.info("No Project Name - Using name of log file")
        if roadway_log_file:
            roadway_changes = Project.read_logfile(roadway_log_file)
        elif roadway_shp_file:
            roadway_changes = gpd.read_file(roadway_shp_file)
            roadway_changes = DataFrame(roadway_changes.drop("geometry", axis=1))
            roadway_changes["model_node_id"] = 0
        elif roadway_csv_file:
            roadway_changes = pd.read_csv(roadway_csv_file)
            roadway_changes["model_node_id"] = 0
        elif roadway_changes:
            pass
        else:
            msg = "No roadway changes given or processed."
            WranglerLogger.info(msg)
            roadway_changes = pd.DataFrame({})

        if base_roadway_network and base_roadway_dir:
            msg = "Method takes only one of 'base_roadway_network' and 'base_roadway_dir' but both given"
            WranglerLogger.error(msg)
            raise ValueError(msg)
        if base_roadway_dir:
            base_roadway_network = load_roadway(
                links_file=os.path.join(base_roadway_dir, "link.json"),
                nodes_file=os.path.join(base_roadway_dir, "node.geojson"),
                shapes_file=os.path.join(base_roadway_dir, "shape.geojson")
            )
            # base_roadway_network = ModelRoadwayNetwork.read(
            #     os.path.join(base_roadway_dir, "link.json"),
            #     os.path.join(base_roadway_dir, "node.geojson"),
            #     os.path.join(base_roadway_dir, "shape.geojson"),
            #     fast=True,
            #     recalculate_calculated_variables=recalculate_calculated_variables,
            #     recalculate_distance=recalculate_distance,
            #     parameters=parameters,
            # )
            base_roadway_network = split_properties_by_time_period_and_category(
                roadway_net=base_roadway_network, 
                parameters=parameters
            )
        elif base_roadway_network:
            # if not isinstance(base_roadway_network, ModelRoadwayNetwork):
            #     base_roadway_network = ModelRoadwayNetwork.from_RoadwayNetwork(
            #         roadway_network_object=base_roadway_network, parameters=parameters
            #     )
            base_roadway_network = split_properties_by_time_period_and_category(
                roadway_net=base_roadway_network, 
                parameters=parameters
            )
        else:
            msg = "No base roadway network."
            WranglerLogger.info(msg)
            base_roadway_network = None

        project = Project(
            roadway_changes=roadway_changes,
            transit_changes=transit_changes,
            base_roadway_network=base_roadway_network,
            base_transit_network=base_transit_network,
            build_transit_network=build_transit_network,
            evaluate=True,
            project_name=project_name,
            parameters=parameters,
            # transit_shape_crosswalk_dict=transit_shape_crosswalk_dict,
        )

        return project

    @staticmethod
    def read_logfile(logfilename: Union[str, List[str]]) -> DataFrame:
        """
        Reads a Cube log file and returns a dataframe of roadway_changes

        Args:
            logfilename (str or list[str]): File path to CUBE logfile or list of logfile paths.

        Returns:
            A DataFrame reprsentation of the log file.
        """
        if type(logfilename) == str:
            logfilename = [logfilename]

        link_df = pd.DataFrame()
        node_df = pd.DataFrame()

        for file in logfilename:
            WranglerLogger.info("Reading logfile: {}".format(file))
            with open(file) as f:
                _content = f.readlines()

                _node_lines = [
                    x.strip().replace(";", ",") for x in _content if x.startswith("N")
                ]
                WranglerLogger.debug("node lines: {}".format(_node_lines))
                _link_lines = [
                    x.strip().replace(";", ",") for x in _content if x.startswith("L")
                ]
                WranglerLogger.debug("link lines: {}".format(_link_lines))

                _nodecol = ["OBJECT", "OPERATION", "GROUP"] + _node_lines[0].split(",")[
                    1:
                ]
                WranglerLogger.debug("Node Cols: {}".format(_nodecol))
                _linkcol = ["OBJECT", "OPERATION", "GROUP"] + _link_lines[0].split(",")[
                    1:
                ]
                WranglerLogger.debug("Link Cols: {}".format(_linkcol))

                def split_log(x):
                    return list(reader([x], delimiter=",", quotechar='"'))[0]

                _node_df = pd.DataFrame(
                    [split_log(x) for x in _node_lines[1:]], columns=_nodecol
                )
                WranglerLogger.debug("Node DF: {}".format(_node_df))
                _link_df = pd.DataFrame(
                    [split_log(x) for x in _link_lines[1:]], columns=_linkcol
                )
                WranglerLogger.debug("Link DF: {}".format(_link_df))

                node_df = pd.concat([node_df, _node_df])
                link_df = pd.concat([link_df, _link_df])

        log_df = pd.concat([link_df, node_df], ignore_index=True, sort=False)

        # CUBE logfile headers for string fields: NAME[111] instead of NAME, need to shorten that
        log_df.columns = [c.split("[")[0] for c in log_df.columns]

        WranglerLogger.info(
            "Processed {} Node lines and {} Link lines".format(
                node_df.shape[0], link_df.shape[0]
            )
        )

        return log_df

    @staticmethod
    def determine_roadway_network_changes_compatibility(
        base_roadway_network,
        roadway_changes: DataFrame,
        parameters: Parameters,
    ):
        """
        Checks to see that any links or nodes that change exist in base roadway network.
        """
        WranglerLogger.info(
            "Evaluating compatibility between roadway network changes and base network. Not evaluating deletions."
        )

        # CUBE log file saves all variable names in upper cases, need to convert them to be same as network
        log_to_net_df = pd.read_csv(parameters.log_to_net_crosswalk)
        log_to_net_dict = dict(zip(log_to_net_df["log"], log_to_net_df["net"]))

        dbf_to_net_df = pd.read_csv(parameters.net_to_dbf_crosswalk)
        dbf_to_net_dict = dict(zip(dbf_to_net_df["dbf"], dbf_to_net_df["net"]))

        roadway_changes.rename(columns=log_to_net_dict, inplace=True)
        roadway_changes.rename(columns=dbf_to_net_dict, inplace=True)

        # for links "L"  that change "C",
        # find locations where there isn't a base roadway link

        link_changes_df = roadway_changes[
            (roadway_changes.OBJECT == "L") & (roadway_changes.OPERATION == "C")
        ].copy()
        link_changes_df["A_B"] = (
            link_changes_df["A"].astype(str) + "_" + link_changes_df["B"].astype(str)
        )

        link_additions_df = roadway_changes[
            (roadway_changes.OBJECT == "L") & (roadway_changes.OPERATION == "A")
        ].copy()

        if len(link_additions_df) > 0:
            link_additions_df["A_B"] = (
                link_additions_df["A"].astype(str)
                + "_"
                + link_additions_df["B"].astype(str)
            )

            link_changes_df = link_changes_df[
                ~(link_changes_df["A_B"].isin(link_additions_df["A_B"].tolist()))
            ].copy()

        link_merge_df = pd.merge(
            link_changes_df[["A", "B"]].astype(str),
            base_roadway_network.links_df[["A", "B", "model_link_id"]].astype(str),
            how="left",
            on=["A", "B"],
        )

        missing_links = link_merge_df.loc[link_merge_df["model_link_id"].isna()]

        if missing_links.shape[0]:
            msg = "Network missing the following AB links:\n{}".format(missing_links)
            WranglerLogger.error(msg)
            raise ValueError(msg)

        # for links "N"  that change "C",
        # find locations where there isn't a base roadway node

        node_changes_df = roadway_changes[
            (roadway_changes.OBJECT == "N") & (roadway_changes.OPERATION == "C")
        ].copy()

        node_additions_df = roadway_changes[
            (roadway_changes.OBJECT == "N") & (roadway_changes.OPERATION == "A")
        ].copy()

        if len(node_additions_df) > 0:
            node_changes_df = node_changes_df[
                ~(
                    node_changes_df["model_node_id"].isin(
                        node_additions_df["model_node_id"].tolist()
                    )
                )
            ]

        node_merge_df = pd.merge(
            node_changes_df[["model_node_id"]],
            base_roadway_network.nodes_df[["model_node_id", "geometry"]],
            how="left",
            on=["model_node_id"],
        )
        missing_nodes = node_merge_df.loc[node_merge_df["geometry"].isna()]
        if missing_nodes.shape[0]:
            msg = "Network missing the following nodes:\n{}".format(missing_nodes)
            WranglerLogger.error(msg)
            raise ValueError(msg)

    def evaluate_changes(self):
        """
        Determines which changes should be evaluated, initiates
        self.card_data to be an aggregation of transit and highway changes.
        """
        highway_change_list = []
        transit_change_list = []

        WranglerLogger.info("Evaluating project changes.")

        if not self.roadway_changes.empty:
            highway_change_list = self.add_highway_changes()

        if (self.transit_changes is not None) or (
            self.base_transit_network is not None
            and self.build_transit_network is not None
        ):
            transit_change_list = self.add_transit_changes()

        self.card_data = {
            "project": self.project_name,
            "changes": transit_change_list + highway_change_list,
        }

    def add_transit_changes(self):
        """
        Evaluates changes between base and build transit objects and
        adds entries into the self.card_data dictionary.
        """

        transit_change_list = self.build_transit_network.evaluate_differences(
            self.base_transit_network
        )

        return transit_change_list

    def add_highway_changes(self, limit_variables_to_existing_network=False):
        """
        Evaluates changes from the log file based on the base highway object and
        adds entries into the self.card_data dictionary.

        Args:
            limit_variables_to_existing_network (bool): True if no ad-hoc variables.  Default to False.
        """

        for c in self.parameters.string_col:
            if c in self.roadway_changes.columns:
                self.roadway_changes[c] = self.roadway_changes[c].str.lstrip(" ")

        ## if worth it, could also add some functionality  to network wrangler itself.
        node_changes_df = self.roadway_changes[
            self.roadway_changes.OBJECT == "N"
        ].copy()

        link_changes_df = self.roadway_changes[
            self.roadway_changes.OBJECT == "L"
        ].copy()

        def _final_op(x):
            if x.OPERATION_history[-1] == "D":
                if "A" in x.OPERATION_history[:-1]:
                    return "N"
                else:
                    return "D"
            elif x.OPERATION_history[-1] == "A":
                if "D" in x.OPERATION_history[:-1]:
                    return "C"
                else:
                    return "A"
            else:
                if "A" in x.OPERATION_history[:-1]:
                    return "A"
                else:
                    return "C"

        def _process_deletions(link_changes_df):
            """ """
            WranglerLogger.debug("Processing link deletions")

            cube_delete_df = link_changes_df[link_changes_df.OPERATION_final == "D"]
            if len(cube_delete_df) > 0:
                links_to_delete = cube_delete_df["model_link_id"].tolist()
                delete_link_dict = {
                    "roadway_deletion":{
                        "links": {
                            "model_link_id": links_to_delete, 
                            "modes": ["any"]
                        }
                    },
                }
                WranglerLogger.debug("{} Links Deleted.".format(len(links_to_delete)))
            else:
                delete_link_dict = None
                WranglerLogger.debug("No link deletions processed")

            return delete_link_dict

        def _process_link_additions(
            link_changes_df, limit_variables_to_existing_network
        ):
            """ """
            WranglerLogger.debug("Processing link additions")
            cube_add_df = link_changes_df[link_changes_df.OPERATION_final == "A"].copy()
            if len(cube_add_df) == 0:
                WranglerLogger.debug("No link additions processed")
                return {}

            if limit_variables_to_existing_network:
                add_col = [
                    c
                    for c in cube_add_df.columns
                    if c in self.base_roadway_network.links_df.columns
                ]
            else:
                add_col = [
                    c for c in cube_add_df.columns if c not in ["OPERATION_final"]
                ]
                # can leave out "OPERATION_final" from writing out, is there a reason to write it out?

            # convert column types
            for bc in list(set(self.parameters.bool_col) & set(cube_add_df.columns)):
                try:
                    cube_add_df[bc] = (
                        pd.to_numeric(cube_add_df[bc]).astype(int).astype(bool)
                    )
                except ValueError:
                    msg = 'Cannot convert column "{}" with values "{}" to boolean.'.format(
                        bc, cube_add_df[bc].unique()
                    )
                    WranglerLogger.error(msg)
                    raise ValueError(msg)
            for bc in list(set(self.parameters.int_col) & set(cube_add_df.columns)):
                cube_add_df[bc] = cube_add_df[bc].astype(int)
            for bc in list(set(self.parameters.float_col) & set(cube_add_df.columns)):
                cube_add_df[bc] = cube_add_df[bc].astype(float)
            for bc in list(set(self.parameters.string_col) & set(cube_add_df.columns)):
                cube_add_df[bc] = cube_add_df[bc].astype(str)

            add_col_final = add_col.copy()
            # properties that are split by time period and category
            for c in add_col:
                if "_" not in c:
                    continue
                # WranglerLogger.debug("Processing Column: {}".format(c))
                (
                    p_base_name,
                    p_time_period,
                    p_category,
                    managed_lane,
                ) = column_name_to_parts(c, self.parameters)

                if p_base_name not in self.parameters.properties_to_split.keys():
                    continue

                if p_base_name in cube_add_df.columns:
                    if cube_add_df[p_base_name].equals(cube_add_df[c]):
                        add_col_final.remove(c)
                        continue
                    else:
                        msg = "Detected different changes for split-property variables on regular roadway links: "
                        msg += 'conflicting "{}" values "{}", "{}"'.format(
                            p_base_name, cube_add_df[p_base_name], cube_add_df[c]
                        )
                        WranglerLogger.error(msg)
                        raise ValueError(msg)

                else:
                    cube_add_df[p_base_name] = cube_add_df[c]
                    add_col_final.remove(c)
                    add_col_final.append(p_base_name)

            add_link_properties = cube_add_df[add_col_final].to_dict("records")

            updated_add_link_dict = {"roadway_addition":{"links":[]}}
            for link_dict in add_link_properties:
                updated_add_link_dict["roadway_addition"]["links"].append(
                    link_dict
                ) 

            # WranglerLogger.debug("Add Link Properties: {}".format(add_link_properties))
            WranglerLogger.debug("{} Links Added".format(len(add_link_properties)))

            return updated_add_link_dict

        def _process_node_additions(node_add_df):
            """ """
            WranglerLogger.debug("Processing node additions")

            if len(node_add_df) == 0:
                WranglerLogger.debug("No node additions processed")
                return []

            WranglerLogger.info(
                "Reprojecting node coordinates from model CRS to standard CRS"
            )
            if "X" in node_add_df.columns:
                node_add_df = gpd.GeoDataFrame(
                    node_add_df,
                    geometry=gpd.points_from_xy(node_add_df["X"], node_add_df["Y"]),
                    crs="EPSG:" + str(self.parameters.output_epsg),
                )

            # convert to WGS 84
            node_add_df = node_add_df.to_crs(epsg=self.parameters.wrangler_epsg)
            node_add_df["X"] = node_add_df["geometry"].apply(lambda g: g.x)
            node_add_df["Y"] = node_add_df["geometry"].apply(lambda g: g.y)

            add_nodes_dict_list = node_add_df.drop(
                ["OPERATION_final", "geometry"], axis=1
            ).to_dict("records")
            WranglerLogger.debug("{} Nodes Added".format(len(add_nodes_dict_list)))

            return add_nodes_dict_list

        def _process_single_link_change(change_row, changeable_col):
            """ """

            #  1. Find associated base year network values
            base_df = self.base_roadway_network.links_df[
                (self.base_roadway_network.links_df["A"] == change_row.A)
                & (self.base_roadway_network.links_df["B"] == change_row.B)
            ].copy()

            if not base_df.shape[0]:
                msg = "No match found in network for AB combination: ({},{}). Incompatible base network.".format(
                    change_row.A, change_row.B
                )
                WranglerLogger.error(msg)
                raise ValueError(msg)

            elif base_df.shape[0] > 1:
                WranglerLogger.warning(
                    "Found more than one match in base network for AB combination: ({},{}). Selecting first one to operate on but AB should be unique to network.".format(
                        row.A, row.B
                    )
                )

            for bc in list(set(self.parameters.bool_col) & set(base_df.columns)):
                base_df[bc] = base_df[bc].astype(bool)

            base_row = base_df.iloc[0]
            # WranglerLogger.debug("Properties with changes: {}".format(changeable_col))

            # 2. find columns that changed (enough)
            changed_col = []
            for col in changeable_col:
                WranglerLogger.debug("Assessing Column: {}".format(col))
                # if it is the same as before, or a static value, don't process as a change
                if isinstance(base_row[col], bool) | isinstance(
                    base_row[col], np.bool_
                ):
                    # when base is boolean, change should be boolean
                    if int(change_row[col]) == base_row[col]:
                        continue
                if isinstance(base_row[col], (int, float)) & isinstance(
                    change_row[col], (int, float)
                ):
                    # when base and change are both numeric, compare as numeric
                    if base_row[col] == change_row[col]:
                        continue
                if str(change_row[col]).strip("\"'").replace(".0", "") == str(
                    base_row[col]
                ).strip("\"'").replace(".0", ""):
                    # when base and change are inconsisteny in type, compare as string
                    continue
                if (col == "roadway_class") & (change_row[col] == 0):
                    continue
                # only look at distance if it has significantly changed
                if col == "distance":
                    if (
                        abs(
                            (change_row[col] - float(base_row[col]))
                            / base_row[col].astype(float)
                        )
                        > 0.01
                    ):
                        changed_col.append(col)
                    else:
                        continue
                else:
                    changed_col.append(col)

            WranglerLogger.debug(
                "Properties with changes that will be processed: {}".format(changed_col)
            )

            if not changed_col:
                return pd.DataFrame()

            # 3. Iterate through columns with changed values and structure the changes as expected in  project card
            property_dict_list = []
            processed_properties = []

            # check if it's a manged lane change
            for c in changed_col:
                if c.startswith("ML_"):
                    # TODO ML project card skeleton
                    msg = "Detected managed lane changes, please create managed lane project card!"
                    WranglerLogger.error(msg)
                    raise ValueError(msg)
                    return

            # regular roadway property change
            for c in changed_col:
                # WranglerLogger.debug("Processing Column: {}".format(c))
                (
                    p_base_name,
                    p_time_period,
                    p_category,
                    managed_lane,
                ) = column_name_to_parts(c, self.parameters)

                # make the type of set value consistent with the base value
                set_value = change_row[c]
                if base_row[c] is not None:
                    if isinstance(base_row[c], int):
                        set_value = int(set_value)
                    elif isinstance(base_row[c], float):
                        set_value = float(set_value)
                    else:
                        None
                    if isinstance(base_row[c], bool):
                        set_value = bool(int(set_value))

                _d = {
                    "existing": base_row[c],
                    "set": set_value,
                }
                if c in Project.CALCULATED_VALUES:
                    _d = {
                        "set": change_row[c],
                    }
                if p_time_period:
                    if managed_lane == 1:
                        _d["timespan"] = list(
                            self.parameters.time_period_to_time[p_time_period]
                        )
                        if p_category:
                            _d["category"] = p_category

                # iterate through existing properties that have been changed and see if you should just add
                if (p_base_name in processed_properties) & (managed_lane == 1):
                    for processed_p in property_dict_list:
                        if processed_p["property"] == p_base_name:
                            processed_p["scoped"] += [_d]
                elif (p_base_name in processed_properties) & (managed_lane == 0):
                    for processed_p in property_dict_list:
                        if processed_p["property"] == p_base_name:
                            if processed_p["set"] != change_row[c]:
                                msg = "Detected different changes for split-property variables on regular roadway links: "
                                msg += 'conflicting "{}" values "{}", "{}"'.format(
                                    p_base_name, processed_p["set"], change_row[c]
                                )
                                WranglerLogger.error(msg)
                                raise ValueError(msg)
                elif p_time_period:
                    if managed_lane == 1:
                        property_dict = {"property": p_base_name, "scoped": [_d]}
                        processed_properties.append(p_base_name)
                        property_dict_list.append(property_dict)
                    else:
                        _d["property"] = p_base_name
                        processed_properties.append(_d["property"])
                        property_dict_list.append(_d)
                else:
                    _d["property"] = p_base_name
                    processed_properties.append(_d["property"])
                    property_dict_list.append(_d)

            card_df = pd.DataFrame(
                {
                    "properties": pd.Series([property_dict_list]),
                    "model_link_id": pd.Series(base_row["model_link_id"]),
                }
            )

            # WranglerLogger.debug('single change card_df:\n {}'.format(card_df))

            return card_df

        def _process_link_changes(link_changes_df, changeable_col):
            """ """
            cube_change_df = link_changes_df[link_changes_df.OPERATION_final == "C"]
            if not cube_change_df.shape[0]:
                WranglerLogger.info("No link changes processed")
                return []

            change_link_dict_df = pd.DataFrame(columns=["properties", "model_link_id"])

            for index, row in cube_change_df.iterrows():
                card_df = _process_single_link_change(row, changeable_col)
                change_link_dict_df = pd.concat(
                    [change_link_dict_df, card_df], ignore_index=True, sort=False
                )

            if not change_link_dict_df.shape[0]:
                WranglerLogger.info("No link changes processed")
                return []

            # WranglerLogger.debug('change_link_dict_df Unaggregated:\n {}'.format(change_link_dict_df))

            # Have to change to string so that it is a hashable type for the aggregation
            change_link_dict_df["properties"] = change_link_dict_df[
                "properties"
            ].astype(str)
            # Group the changes that are the same
            change_link_dict_df = (
                change_link_dict_df.groupby("properties")[["model_link_id"]]
                .agg(lambda x: list(x))
                .reset_index()
            )
            # WranglerLogger.debug('change_link_dict_df Aggregated:\n {}'.format(change_link_dict_df))

            # Reformat model link id to correct "facility" format
            change_link_dict_df["facility"] = change_link_dict_df.apply(
                lambda x: {"links": {"model_link_id": x.model_link_id}}, axis=1
            )

            # WranglerLogger.debug('change_link_dict_df 3: {}'.format(change_link_dict_df))
            change_link_dict_df["properties"] = change_link_dict_df["properties"].apply(
                lambda x: json.loads(
                    x.replace("'\"", "'")
                    .replace("\"'", "'")
                    .replace("'", '"')
                    .replace("False", "false")
                    .replace("True", "true")
                )
            )

            # change_link_dict_df["category"] = "Roadway Property Change"

            change_link_dict_list = change_link_dict_df[
                ["facility", "properties"]
            ].to_dict("records")

            # update the change link dictionry to be consistent with the new project card schema
            updated_change_link_dict_list = []
            for link_dict in change_link_dict_list:
                updated_properties = {
                    item.pop("property"): item
                    for item in link_dict["properties"]
                }
                link_dict["facility"]["links"].update({"modes": ["any"]})
                updated_link_dict = {
                    "roadway_property_change":{
                        "property_changes": updated_properties,
                        "facility": link_dict["facility"]
                    }
                }
                updated_change_link_dict_list.append(updated_link_dict)

            WranglerLogger.debug(
                "{} Changes Processed".format(len(updated_change_link_dict_list))
            )
            return updated_change_link_dict_list

        def _consolidate_actions(log, base, key_list):
            log_df = log.copy()
            # will be changed if to allow new variables being added/changed that are not in base network
            changeable_col = [x for x in log_df.columns if x in base.columns]

            for x in changeable_col:
                log_df[x] = log_df[x].astype(base[x].dtype)

            action_history_df = (
                log_df.groupby(key_list)["OPERATION"]
                .agg(lambda x: x.tolist())
                .rename("OPERATION_history")
                .reset_index()
            )

            log_df = pd.merge(log_df, action_history_df, on=key_list, how="left")
            log_df.drop_duplicates(subset=key_list, keep="last", inplace=True)
            log_df["OPERATION_final"] = log_df.apply(lambda x: _final_op(x), axis=1)
            return log_df[changeable_col + ["OPERATION_final"]]

        delete_link_dict = None
        add_link_dict = []
        change_link_dict_list = []

        if len(link_changes_df) != 0:
            link_changes_df = _consolidate_actions(
                link_changes_df, self.base_roadway_network.links_df, ["A", "B"]
            )

            # process deletions
            delete_link_dict = _process_deletions(link_changes_df)

            # process additions
            add_link_dict = _process_link_additions(
                link_changes_df, limit_variables_to_existing_network
            )

            # process changes
            WranglerLogger.debug("Processing changes")
            WranglerLogger.debug(link_changes_df)
            changeable_col = list(
                (
                    set(link_changes_df.columns)
                    & set(self.base_roadway_network.links_df.columns)
                )
                - set(Project.STATIC_VALUES)
            )

            cols_in_changes_not_in_net = list(
                set(link_changes_df.columns)
                - set(self.base_roadway_network.links_df.columns)
            )

            if cols_in_changes_not_in_net:
                WranglerLogger.warning(
                    "The following attributes are specified in the changes but do not exist in the base network: {}".format(
                        cols_in_changes_not_in_net
                    )
                )

            for c in changeable_col:
                if c in self.parameters.string_col:
                    self.base_roadway_network.links_df[c].fillna("", inplace=True)
                elif c in self.parameters.bool_col:
                    self.base_roadway_network.links_df[c].fillna(False, inplace=True)
                else:
                    self.base_roadway_network.links_df[c].fillna(0, inplace=True)

            change_link_dict_list = _process_link_changes(
                link_changes_df, changeable_col
            )

        if len(node_changes_df) != 0:
            node_changes_df = _consolidate_actions(
                node_changes_df, self.base_roadway_network.nodes_df, ["model_node_id"]
            )

            # print error message for node change and node deletion
            if (
                len(node_changes_df[node_changes_df.OPERATION_final.isin(["C", "D"])])
                > 0
            ):
                msg = "NODE changes and deletions are not allowed!"
                WranglerLogger.error(msg)
                raise ValueError(msg)
            node_add_df = node_changes_df[node_changes_df.OPERATION_final == "A"]

            if add_link_dict:
                add_link_dict["roadway_addition"].update(
                    {"nodes": _process_node_additions(node_add_df)}
                )
            else:
                add_link_dict.append(
                    {
                        "roadway_addition":{
                            "nodes": _process_node_additions(node_add_df),
                        },
                    }
                )

        else:
            None

        # combine together

        highway_change_list = list(
            filter(None, [delete_link_dict] + [add_link_dict] + change_link_dict_list)
        )

        return highway_change_list



__version__ = "0.0.0"

from .project import Project
from .transit import StandardTransit
from .util import get_shared_streets_intersection_hash
# from .roadway import ModelRoadwayNetwork
from .parameters import Parameters



import logging

__all__ = ["WranglerLogger", "setupLogging"]


# for all the Wrangler logging needs!
WranglerLogger = logging.getLogger("WranglerLogger")


def setupLogging(infoLogFilename, debugLogFilename, logToConsole=True):
    """Sets up the logger.  The infoLog is terse, just gives the bare minimum of details
    so the network composition will be clear later.
    The debuglog is very noisy, for debugging.

    Pass none to either.
    Spews it all out to console too, if logToConsole is true.
    """
    # clear handlers if any exist already
    WranglerLogger.handlers = []

    # create a logger
    WranglerLogger.setLevel(logging.DEBUG)

    if infoLogFilename:
        infologhandler = logging.StreamHandler(open(infoLogFilename, "w"))
        infologhandler.setLevel(logging.INFO)
        infologhandler.setFormatter(
            logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        )
        WranglerLogger.addHandler(infologhandler)

    if debugLogFilename:
        debugloghandler = logging.StreamHandler(open(debugLogFilename, "w"))
        debugloghandler.setLevel(logging.DEBUG)
        debugloghandler.setFormatter(
            logging.Formatter("%(asctime)s %(levelname)s %(message)s", "%Y-%m-%d %H:%M")
        )
        WranglerLogger.addHandler(debugloghandler)

    if logToConsole:
        consolehandler = logging.StreamHandler()
        consolehandler.setLevel(logging.DEBUG)
        consolehandler.setFormatter(
            logging.Formatter("%(name)-12s: %(levelname)-8s %(message)s")
        )
        WranglerLogger.addHandler(consolehandler)



import copy
from typing import Optional, Union

import geopandas as gpd
import pandas as pd

from geopandas import GeoDataFrame
from pandas import DataFrame
import numpy as np

from network_wrangler.roadway.network import RoadwayNetwork
from network_wrangler.roadway.links.scopes import prop_for_scope

from .parameters import Parameters
from .logger import WranglerLogger


# class ModelRoadwayNetwork(RoadwayNetwork):
#     """
#     Subclass of network_wrangler class :ref:`RoadwayNetwork <network_wrangler:RoadwayNetwork>`

#     A representation of the physical roadway network and its properties.
#     """

#     def __init__(
#         self,
#         nodes: GeoDataFrame,
#         links: DataFrame,
#         shapes: GeoDataFrame,
#         parameters: Union[Parameters, dict] = {},
#     ):
#         """
#         Constructor

#         Args:
#             nodes: geodataframe of nodes
#             links: dataframe of links
#             shapes: geodataframe of shapes
#             parameters: dictionary of parameter settings (see Parameters class) or an instance of Parameters. If not specified, will use default parameters.

#         """
#         super().__init__(nodes, links, shapes)

#         # will have to change if want to alter them
#         if type(parameters) is dict:
#             self.parameters = Parameters(**parameters)
#         elif isinstance(parameters, Parameters):
#             # self.parameters = Parameters(**parameters.__dict__)
#             self.parameters = parameters
#         else:
#             msg = "Parameters should be a dict or instance of Parameters: found {} which is of type:{}".format(
#                 parameters, type(parameters)
#             )
#             WranglerLogger.error(msg)
#             raise ValueError(msg)

#     @staticmethod
#     def read(
#         link_file: str,
#         node_file: str,
#         shape_file: str,
#         fast: bool = False,
#         recalculate_calculated_variables=False,
#         parameters: Union[dict, Parameters] = {},
#     ):
#         """
#         Reads in links and nodes network standard.

#         Args:
#             link_file (str): File path to link json.
#             node_file (str): File path to node geojson.
#             shape_file (str): File path to link true shape geojson
#             fast (bool): boolean that will skip validation to speed up read time.
#             recalculate_calculated_variables (bool): calculates fields from spatial joins, etc.
#             recalculate_distance (bool):  re-calculates distance.
#             parameters: dictionary of parameter settings (see Parameters class) or an instance of Parameters. If not specified, will use default parameters.

#         Returns:
#             ModelRoadwayNetwork
#         """
#         # road_net =  super().read(link_file, node_file, shape_file, fast=fast)
#         road_net = RoadwayNetwork.read(link_file, node_file, shape_file, fast=fast)

#         m_road_net = ModelRoadwayNetwork(
#             road_net.nodes_df,
#             road_net.links_df,
#             road_net.shapes_df,
#             parameters=parameters,
#         )

#         if recalculate_calculated_variables:
#             m_road_net.create_calculated_variables()

#         return m_road_net

#     @staticmethod
#     def from_RoadwayNetwork(
#         roadway_network_object, parameters: Union[dict, Parameters] = {}
#     ):
#         """
#         RoadwayNetwork to ModelRoadwayNetwork

#         Args:
#             roadway_network_object (RoadwayNetwork).
#             parameters: dictionary of parameter settings (see Parameters class) or an instance of Parameters. If not specified, will use default parameters.

#         Returns:
#             ModelRoadwayNetwork
#         """
#         return ModelRoadwayNetwork(
#             roadway_network_object.nodes_df,
#             roadway_network_object.links_df,
#             roadway_network_object.shapes_df,
#             parameters=parameters,
#         )

def split_properties_by_time_period_and_category(roadway_net=None, parameters=None, properties_to_split=None):
    """
    Splits properties by time period, assuming a variable structure of

    Args:
        properties_to_split: dict
            dictionary of output variable prefix mapped to the source variable and what to stratify it by
            e.g.
            {
                'trn_priority' : {'v':'trn_priority', 'times_periods':{"AM": ("6:00", "9:00"),"PM": ("16:00", "19:00")}},
                'ttime_assert' : {'v':'ttime_assert', 'times_periods':{"AM": ("6:00", "9:00"),"PM": ("16:00", "19:00")}},
                'lanes' : {'v':'lanes', 'times_periods':{"AM": ("6:00", "9:00"),"PM": ("16:00", "19:00")}},
                'ML_lanes' : {'v':'ML_lanes', 'times_periods':{"AM": ("6:00", "9:00"),"PM": ("16:00", "19:00")}},
                'price' : {'v':'price', 'times_periods':{"AM": ("6:00", "9:00"),"PM": ("16:00", "19:00")}},'categories': {"sov": ["sov", "default"],"hov2": ["hov2", "default", "sov"]}},
                'access' : {'v':'access', 'times_periods':{"AM": ("6:00", "9:00"),"PM": ("16:00", "19:00")}},
            }

    """
    import itertools

    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)

    if properties_to_split == None:
        properties_to_split = parameters.properties_to_split

    for out_var, params in properties_to_split.items():
        if params["v"] not in roadway_net.links_df.columns:
            WranglerLogger.warning(
                "Specified variable to split: {} not in network variables: {}. Returning 0.".format(
                    params["v"], str(roadway_net.links_df.columns)
                )
            )
            if params.get("time_periods") and params.get("categories"):

                for time_suffix, category_suffix in itertools.product(
                    params["time_periods"], params["categories"]
                ):
                    roadway_net.links_df[
                        out_var + "_" + time_suffix + "_" + category_suffix
                    ] = 0
            elif params.get("time_periods"):
                for time_suffix in params["time_periods"]:
                    roadway_net.links_df[out_var + "_" + time_suffix] = 0
        elif params.get("time_periods") and params.get("categories"):
            for time_suffix, category_suffix in itertools.product(
                params["time_periods"], params["categories"]
            ):
                roadway_net.links_df[
                    out_var + "_" + category_suffix + "_" + time_suffix
                ] = prop_for_scope(
                    roadway_net.links_df,
                    params["v"],
                    category=params["categories"][category_suffix],
                    timespan=params["time_periods"][time_suffix],
                )[params["v"]]
        elif params.get("time_periods"):
            for time_suffix in params["time_periods"]:
                roadway_net.links_df[
                    out_var + "_" + time_suffix
                ] = prop_for_scope(
                    roadway_net.links_df,
                    params["v"],
                    category=None,
                    timespan=params["time_periods"][time_suffix],
                )[params["v"]]
        else:
            raise ValueError(
                "Shoudn't have a category without a time period: {}".format(params)
            )

    roadway_net.links_df.attrs = link_attrs

    return roadway_net
    
# def create_calculated_variables(self):
#     """
#     Creates calculated roadway variables.

#     Args:
#         None
#     """
#     WranglerLogger.info("Creating calculated roadway variables.")
#     self.calculate_distance_miles(overwrite=True)

def calculate_distance_miles(roadway_net=None, network_variable="distance", overwrite=False):
    """
    calculate link distance in miles

    Args:
        overwrite (Bool): True if overwriting existing variable in network.  Default to False.

    Returns:
        None

    """

    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)

    if network_variable in roadway_net.links_df:
        if overwrite or (roadway_net.links_df[network_variable].isnull().any()):
            WranglerLogger.info(
                "Overwriting existing distance Variable '{}' already in network".format(
                    network_variable
                )
            )
        else:
            WranglerLogger.info(
                "Distance Variable '{}' already in network. Returning without overwriting.".format(
                    network_variable
                )
            )
            return roadway_net

    """
    Start actual process
    """

    temp_links_gdf = roadway_net.links_df.copy()
    temp_links_gdf.crs = "EPSG:4326"
    temp_links_gdf = temp_links_gdf.to_crs(epsg=26915)

    WranglerLogger.info(
        "Calculating distance in miles for all links".format(network_variable)
    )
    temp_links_gdf[network_variable] = temp_links_gdf.geometry.length / 1609.34
    # overwrite 0 distance with 0.001 mile
    temp_links_gdf[network_variable] = np.where(
        temp_links_gdf[network_variable] == 0,
        0.001,
        temp_links_gdf[network_variable],
    )

    roadway_net.links_df[network_variable] = temp_links_gdf[network_variable]
    roadway_net.links_df.attrs = link_attrs

    return roadway_net

def calculate_distance(
    roadway_net=None, network_variable="distance", centroidconnect_only=False, overwrite=False
):
    """
    calculate link distance in miles

    Args:
        centroidconnect_only (Bool):  True if calculating distance for centroidconnectors only.  Default to True.
        overwrite (Bool): True if overwriting existing variable in network.  Default to False.

    Returns:
        None

    """

    if network_variable in roadway_net.links_df:
        if overwrite:
            WranglerLogger.info(
                "Overwriting existing distance Variable '{}' already in network".format(
                    network_variable
                )
            )
        else:
            WranglerLogger.info(
                "Distance Variable '{}' already in network. Returning without overwriting.".format(
                    network_variable
                )
            )
            return roadway_net

    """
    Verify inputs
    """

    if ("centroidconnect" not in roadway_net.links_df) & (
        "taz" not in roadway_net.links_df.roadway.unique()
    ):
        if centroidconnect_only:
            msg = "No variable specified for centroid connector, calculating centroidconnect first"
            WranglerLogger.error(msg)
            raise ValueError(msg)

    """
    Start actual process
    """
    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)

    temp_links_gdf = roadway_net.links_df.copy()
    temp_links_gdf.crs = "EPSG:4326"
    temp_links_gdf = temp_links_gdf.to_crs(epsg=26915)

    if centroidconnect_only:
        WranglerLogger.info(
            "Calculating {} for centroid connectors".format(network_variable)
        )
        temp_links_gdf[network_variable] = np.where(
            temp_links_gdf.centroidconnect == 1,
            temp_links_gdf.geometry.length / 1609.34,
            temp_links_gdf[network_variable],
        )
    else:
        WranglerLogger.info(
            "Calculating distance for all links".format(network_variable)
        )
        temp_links_gdf[network_variable] = temp_links_gdf.geometry.length / 1609.34
        # overwrite 0 distance with 0.001 mile
        temp_links_gdf.loc[
            temp_links_gdf[network_variable] == 0, network_variable
        ] = 0.001

    roadway_net.links_df[network_variable] = temp_links_gdf[network_variable]
    roadway_net.links_df.attrs = link_attrs

    return roadway_net

def create_ML_variable(
    roadway_net=None,
    network_variable="ML_lanes",
    overwrite=False,
):
    """
    Created ML lanes placeholder for project to write out ML changes

    ML lanes default to 0, ML info comes from cube LOG file and store in project cards

    Args:
        overwrite (Bool): True if overwriting existing variable in network.  Default to False.

    Returns:
        None
    """
    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)

    if network_variable in roadway_net.links_df:
        if overwrite:
            WranglerLogger.info(
                "Overwriting existing ML Variable '{}' already in network".format(
                    network_variable
                )
            )
            roadway_net.links_df[network_variable] = int(0)
        else:
            WranglerLogger.info(
                "ML Variable '{}' already in network. Returning without overwriting.".format(
                    network_variable
                )
            )
            return roadway_net

    """
    Verify inputs
    """

    WranglerLogger.info(
        "Finished creating ML lanes variable: {}".format(network_variable)
    )
    roadway_net.links_df.attrs = link_attrs

    return roadway_net

def create_hov_corridor_variable(
    roadway_net=None,
    network_variable="segment_id",
    overwrite=False,
):
    """
    Created hov corridor placeholder for project to write out corridor changes

    hov corridor id default to 0, its info comes from cube LOG file and store in project cards

    Args:
        overwrite (Bool): True if overwriting existing variable in network.  Default to False.

    Returns:
        None
    """
    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)

    if network_variable in roadway_net.links_df:
        if overwrite:
            WranglerLogger.info(
                "Overwriting existing hov corridor Variable '{}' already in network".format(
                    network_variable
                )
            )
        else:
            WranglerLogger.info(
                "Hov corridor Variable '{}' already in network. Returning without overwriting.".format(
                    network_variable
                )
            )
            return roadway_net

    """
    Verify inputs
    """

    roadway_net.links_df[network_variable] = int(0)

    WranglerLogger.info(
        "Finished creating hov corridor variable: {}".format(network_variable)
    )
    roadway_net.links_df.attrs = link_attrs

    return roadway_net

def create_managed_variable(
    roadway_net=None,
    network_variable="managed",
    overwrite=False,
):
    """
    Created placeholder for project to write out managed

    managed default to 0, its info comes from cube LOG file and store in project cards

    Args:
        overwrite (Bool): True if overwriting existing variable in network.  Default to False.

    Returns:
        None
    """
    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)

    if network_variable in roadway_net.links_df:
        if overwrite:
            WranglerLogger.info(
                "Overwriting existing managed Variable '{}' already in network".format(
                    network_variable
                )
            )
        else:
            WranglerLogger.info(
                "Managed Variable '{}' already in network. Returning without overwriting.".format(
                    network_variable
                )
            )
            return roadway_net

    """
    Verify inputs
    """

    roadway_net.links_df[network_variable] = int(0)

    WranglerLogger.info(
        "Finished creating managed variable: {}".format(network_variable)
    )
    roadway_net.links_df.attrs = link_attrs

    return roadway_net

def add_variable_using_shst_reference(
    roadway_net=None,
    var_shst_csvdata=None,
    shst_csv_variable=None,
    network_variable=None,
    network_var_type=int,
    overwrite=False,
):
    """
    Join network links with source data, via SHST API node match result.

    Args:
        var_shst_csvdata (str): File path to SHST API return.
        shst_csv_variable (str): Variable name in the source data.
        network_variable (str): Name of the variable that should be written to.
        network_var_type : Variable type in the written network.
        overwrite (bool): True is overwriting existing variable. Default to False.

    Returns:
        None

    """
    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)

    WranglerLogger.info(
        "Adding Variable {} using Shared Streets Reference from {}".format(
            network_variable, var_shst_csvdata
        )
    )

    var_shst_df = pd.read_csv(var_shst_csvdata)
    # there are aadt = 0 in the counts, drop them
    var_shst_df = var_shst_df[var_shst_df[shst_csv_variable] > 0].copy()
    # count station to shared street match - there are many-to-one matches, keep just one match
    var_shst_df.drop_duplicates(subset=["shstReferenceId"], inplace=True)

    if "shstReferenceId" not in var_shst_df.columns:
        msg = "'shstReferenceId' required but not found in {}".format(var_shst_data)
        WranglerLogger.error(msg)
        raise ValueError(msg)

    if shst_csv_variable not in var_shst_df.columns:
        msg = "{} required but not found in {}".format(
            shst_csv_variable, var_shst_data
        )
        WranglerLogger.error(msg)
        raise ValueError(msg)

    join_gdf = pd.merge(
        roadway_net.links_df,
        var_shst_df[["shstReferenceId", shst_csv_variable]],
        how="left",
        on="shstReferenceId",
    )

    join_gdf[shst_csv_variable].fillna(0, inplace=True)

    if network_variable in roadway_net.links_df.columns and not overwrite:
        join_gdf.loc[join_gdf[network_variable] == 0, network_variable] = join_gdf[
            shst_csv_variable
        ].astype(network_var_type)
    else:
        join_gdf[network_variable] = join_gdf[shst_csv_variable].astype(
            network_var_type
        )

    roadway_net.links_df[network_variable] = join_gdf[network_variable]

    # MN and WI counts are vehicles using the segment in both directions, no directional counts
    # we will make sure both direction has the same daily AADT
    dir_link_count_df = roadway_net.links_df[
        (roadway_net.links_df[network_variable] > 0) & (roadway_net.links_df["drive_access"] == True)
    ][["A", "B", network_variable]].copy()
    reverse_dir_link_count_df = dir_link_count_df.rename(
        columns={"A": "B", "B": "A"}
    ).copy()

    link_count_df = pd.concat(
        [dir_link_count_df, reverse_dir_link_count_df],
        sort=False,
        ignore_index=True,
    )
    link_count_df.drop_duplicates(subset=["A", "B"], inplace=True)

    roadway_net.links_df = pd.merge(
        roadway_net.links_df.drop(network_variable, axis=1),
        link_count_df[["A", "B", network_variable]],
        how="left",
        on=["A", "B"],
    )
    roadway_net.links_df[network_variable].fillna(0, inplace=True)
    WranglerLogger.info(
        "Added variable: {} using Shared Streets Reference".format(network_variable)
    )
    roadway_net.links_df.attrs = link_attrs

    return roadway_net

def convert_int(roadway_net=None, parameters=None, int_col_names=[]):
    """
    Convert integer columns
    """
    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)
    node_attrs = copy.deepcopy(roadway_net.nodes_df.attrs)

    WranglerLogger.info("Converting variable type to MetCouncil standard")

    if not int_col_names:
        int_col_names = parameters.int_col

    ##Why are we doing this?
    # int_col_names.remove("lanes")

    for c in list(set(roadway_net.links_df.columns) & set(int_col_names)):
        roadway_net.links_df[c] = roadway_net.links_df[c].replace(np.nan, 0)
        # REPLACE BLANKS WITH ZERO FOR INTEGER COLUMNS
        roadway_net.links_df[c] = roadway_net.links_df[c].replace("", 0)
        try:
            roadway_net.links_df[c] = roadway_net.links_df[c].astype(int)
        except ValueError:
            try:
                roadway_net.links_df[c] = roadway_net.links_df[c].astype(float)
                roadway_net.links_df[c] = roadway_net.links_df[c].astype(int)
            except:
                msg = f"Could not convert column {c} to integer."
                WranglerLogger.error(msg)
                raise ValueError(msg)

    for c in list(set(roadway_net.nodes_df.columns) & set(int_col_names)):
        roadway_net.nodes_df[c] = roadway_net.nodes_df[c].replace(np.nan, 0)
        # REPLACE BLANKS WITH ZERO FOR INTEGER COLUMNS
        roadway_net.nodes_df[c] = roadway_net.nodes_df[c].replace("", 0)
        try:
            roadway_net.nodes_df[c] = roadway_net.nodes_df[c].astype(int)
        except ValueError:
            msg = f"Could not convert column {c} to integer."
            WranglerLogger.error(msg)
            raise ValueError(msg)

    roadway_net.links_df.attrs = link_attrs
    roadway_net.nodes_df.attrs = node_attrs

    return roadway_net

def convert_bool(roadway_net=None, parameters=None, bool_col_names=[]):
    """
    Convert boolean columns
    """
    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)
    node_attrs = copy.deepcopy(roadway_net.nodes_df.attrs)

    WranglerLogger.info("Converting variable type to MetCouncil standard")

    if not bool_col_names:
        bool_col_names = parameters.bool_col_names

    for c in list(set(roadway_net.links_df.columns) & set(bool_col_names)):
        roadway_net.links_df[c] = roadway_net.links_df[c].replace(np.nan, False)
        # REPLACE BLANKS WITH ZERO FOR INTEGER COLUMNS
        roadway_net.links_df[c] = roadway_net.links_df[c].replace("", False)
        roadway_net.links_df[c] = roadway_net.links_df[c].replace("0", False)
        roadway_net.links_df[c] = roadway_net.links_df[c].replace("1", True)
        try:
            roadway_net.links_df[c] = roadway_net.links_df[c].astype(bool)
        except ValueError:
            msg = f"Could not convert column {c} to boolean."
            WranglerLogger.error(msg)
            raise ValueError(msg)

    for c in list(set(roadway_net.nodes_df.columns) & set(bool_col_names)):
        roadway_net.nodes_df[c] = roadway_net.nodes_df[c].replace(np.nan, False)
        roadway_net.nodes_df[c] = roadway_net.nodes_df[c].replace("", False)
        roadway_net.nodes_df[c] = roadway_net.nodes_df[c].replace("0", False)
        roadway_net.nodes_df[c] = roadway_net.nodes_df[c].replace("1", True)
        try:
            roadway_net.nodes_df[c] = roadway_net.nodes_df[c].astype(bool)
        except ValueError:
            msg = f"Could not convert column {c} to boolean."
            WranglerLogger.error(msg)
            raise ValueError(msg)

    roadway_net.links_df.attrs = link_attrs
    roadway_net.nodes_df.attrs = node_attrs

    return roadway_net


def fill_na(roadway_net=None, parameters=None):
    """
    Fill na values from create_managed_lane_network()
    """
    link_attrs = copy.deepcopy(roadway_net.links_df.attrs)
    node_attrs = copy.deepcopy(roadway_net.nodes_df.attrs)

    WranglerLogger.info("Filling nan for network from network wrangler")

    num_col = parameters.int_col + parameters.float_col

    for x in list(roadway_net.links_df.columns):
        if x in num_col:
            roadway_net.links_df[x].fillna(0, inplace=True)
            roadway_net.links_df[x] = roadway_net.links_df[x].apply(
                lambda k: 0 if k in [np.nan, "", float("nan"), "NaN"] else k
            )

        # else:
        #     roadway_net.links_df[x].fillna("", inplace=True)

    for x in list(roadway_net.nodes_df.columns):
        if x in num_col:
            roadway_net.nodes_df[x].fillna(0, inplace=True)
        else:
            roadway_net.nodes_df[x].fillna("", inplace=True)

    roadway_net.links_df.attrs = link_attrs
    roadway_net.nodes_df.attrs = node_attrs

    return roadway_net

def rename_variables_for_dbf(
    input_df=None,
    parameters=None,
    variable_crosswalk: str = None,
    output_variables: list = None,
    convert_geometry_to_xy=False,
):
    """
    Rename attributes for DBF/SHP, make sure length within 10 chars.

    Args:
        input_df (dataframe): Network standard DataFrame.
        variable_crosswalk (str): File path to variable name crosswalk from network standard to DBF names.
        output_variables (list): List of strings for DBF variables.
        convert_geometry_to_xy (bool): True if converting node geometry to X/Y

    Returns:
        dataframe

    """
    WranglerLogger.info("Renaming variables so that they are DBF-safe")

    """
    Verify inputs
    """

    variable_crosswalk = (
        variable_crosswalk
        if variable_crosswalk
        else parameters.net_to_dbf_crosswalk
    )

    output_variables = (
        output_variables if output_variables else parameters.output_variables
    )

    """
    Start actual process
    """

    crosswalk_df = pd.read_csv(variable_crosswalk)
    WranglerLogger.debug(
        "Variable crosswalk: {} \n {}".format(variable_crosswalk, crosswalk_df)
    )
    net_to_dbf_dict = dict(zip(crosswalk_df["net"], crosswalk_df["dbf"]))

    dbf_name_list = []

    dbf_df = copy.deepcopy(input_df)

    # only write out variables that we specify
    # if variable is specified in the crosswalk, rename it to that variable
    for c in dbf_df.columns:
        if c in output_variables:
            try:
                dbf_df.rename(columns={c: net_to_dbf_dict[c]}, inplace=True)
                dbf_name_list += [net_to_dbf_dict[c]]
            except:
                dbf_name_list += [c]

    if "geometry" in dbf_df.columns:
        if str(dbf_df["geometry"].iloc[0].geom_type) == "Point":
            dbf_df["X"] = dbf_df.geometry.apply(lambda g: g.x)
            dbf_df["Y"] = dbf_df.geometry.apply(lambda g: g.y)
            if "X" not in dbf_name_list:
                dbf_name_list += ["X", "Y"]

    WranglerLogger.debug("DBF Variables: {}".format(",".join(dbf_name_list)))

    return dbf_df[dbf_name_list]

def write_roadway_as_shp(
    roadway_net=None,
    parameters=None,
    node_output_variables: list = None,
    link_output_variables: list = None,
    data_to_csv: bool = True,
    data_to_dbf: bool = False,
    output_link_shp: str = None,
    output_node_shp: str = None,
    output_link_csv: str = None,
    output_node_csv: str = None,
    export_drive_only: bool = False,
):
    """
    Write out dbf/shp for cube.  Write out csv in addition to shp with full length variable names.

    Args:
        node_output_variables (list): List of strings for node output variables.
        link_output_variables (list): List of strings for link output variables.
        data_to_csv (bool): True if write network in csv format.
        data_to_dbf (bool): True if write network in dbf/shp format.
        output_link_shp (str): File path to output link dbf/shp.
        output_node_shp (str): File path to output node dbf/shp.
        output_link_csv (str): File path to output link csv.
        output_node_csv (str): File path to output node csv.
        export_drive_only (bool) : True if write out drive links/nodes only.

    Returns:
        None
    """

    WranglerLogger.info("Writing Network as Shapefile")
    WranglerLogger.debug(
        "Output Variables: \n - {}".format(
            "\n - ".join(parameters.output_variables)
        )
    )

    """
    Verify inputs
    """

    WranglerLogger.debug(
        "Network Link Variables: \n - {}".format(
            "\n - ".join(roadway_net.links_df.columns)
        )
    )
    WranglerLogger.debug(
        "Network Node Variables: \n - {}".format(
            "\n - ".join(roadway_net.nodes_df.columns)
        )
    )

    link_output_variables = (
        link_output_variables
        if link_output_variables
        else [
            c
            for c in roadway_net.links_df.columns
            if c in parameters.output_variables
        ]
    )

    node_output_variables = (
        node_output_variables
        if node_output_variables
        else [
            c
            for c in roadway_net.nodes_df.columns
            if c in parameters.output_variables
        ]
    )

    # unless specified that all the data goes to the DBF, only output A and B
    dbf_link_output_variables = (
        link_output_variables if data_to_dbf else ["A", "B", "shape_id", "geometry"]
    )

    output_link_shp = (
        output_link_shp if output_link_shp else parameters.output_link_shp
    )

    output_node_shp = (
        output_node_shp if output_node_shp else parameters.output_node_shp
    )

    output_link_csv = (
        output_link_csv if output_link_csv else parameters.output_link_csv
    )

    output_node_csv = (
        output_node_csv if output_node_csv else parameters.output_node_csv
    )

    """
    Start Process
    """

    WranglerLogger.info("Renaming DBF Node Variables")
    nodes_dbf_df = rename_variables_for_dbf(
        input_df = roadway_net.nodes_df, 
        parameters = parameters,
        output_variables=node_output_variables
    )
    WranglerLogger.info("Renaming DBF Link Variables")
    links_dbf_df = rename_variables_for_dbf(
        input_df = roadway_net.links_df,
        parameters = parameters,
        output_variables=dbf_link_output_variables
    )

    links_dbf_df = gpd.GeoDataFrame(links_dbf_df, geometry=links_dbf_df["geometry"])

    if export_drive_only == True:
        nodes_dbf_df = nodes_dbf_df[nodes_dbf_df.drive_node == 1].copy()
        links_dbf_df = links_dbf_df[links_dbf_df.drive == 1].copy()

    WranglerLogger.info("Writing Node Shapes:\n - {}".format(output_node_shp))
    nodes_dbf_df.to_file(output_node_shp)
    WranglerLogger.info("Writing Link Shapes:\n - {}".format(output_link_shp))
    links_dbf_df.to_file(output_link_shp)

    if data_to_csv:
        WranglerLogger.info(
            "Writing Network Data to CSVs:\n - {}\n - {}".format(
                output_link_csv, output_node_csv
            )
        )
        roadway_net.links_df[link_output_variables].to_csv(output_link_csv, index=False)
        roadway_net.nodes_df[node_output_variables].to_csv(output_node_csv, index=False)

def write_roadway_as_fixedwidth(
    roadway_net=None,
    parameters=None,
    zones: int=None,
    node_output_variables: list = None,
    link_output_variables: list = None,
    output_link_txt: str = None,
    output_node_txt: str = None,
    output_link_header_width_txt: str = None,
    output_node_header_width_txt: str = None,
    output_cube_network_script: str = None,
    drive_only: bool = False,
):
    """
    Writes out fixed width file.

    This function does:
    1. write out link and node fixed width data files for cube.
    2. write out header and width correspondence.
    3. write out cube network building script with header and width specification.

    Args:
        node_output_variables (list): list of node variable names.
        link_output_variables (list): list of link variable names.
        output_link_txt (str): File path to output link database.
        output_node_txt (str): File path to output node database.
        output_link_header_width_txt (str): File path to link column width records.
        output_node_header_width_txt (str): File path to node column width records.
        output_cube_network_script (str): File path to CUBE network building script.
        drive_only (bool): If True, only writes drive nodes and links

    Returns:
        None

    """

    """
    Verify inputs
    """

    WranglerLogger.debug(
        "Network Link Variables: \n - {}".format(
            "\n - ".join(roadway_net.links_df.columns)
        )
    )
    WranglerLogger.debug(
        "Network Node Variables: \n - {}".format(
            "\n - ".join(roadway_net.nodes_df.columns)
        )
    )

    zones = zones if zones else parameters.zones

    link_output_variables = (
        link_output_variables
        if link_output_variables
        else [
            c
            for c in parameters.output_variables
            if c in roadway_net.links_df.columns
        ]
    )

    node_output_variables = (
        node_output_variables
        if node_output_variables
        else [
            c
            for c in parameters.output_variables
            if c in roadway_net.nodes_df.columns
        ]
    )

    output_link_txt = (
        output_link_txt if output_link_txt else parameters.output_link_txt
    )

    output_node_txt = (
        output_node_txt if output_node_txt else parameters.output_node_txt
    )

    output_link_header_width_txt = (
        output_link_header_width_txt
        if output_link_header_width_txt
        else parameters.output_link_header_width_txt
    )

    output_node_header_width_txt = (
        output_node_header_width_txt
        if output_node_header_width_txt
        else parameters.output_node_header_width_txt
    )

    output_cube_network_script = (
        output_cube_network_script
        if output_cube_network_script
        else parameters.output_cube_network_script
    )

    """
    Start Process
    """
    # convert boolean columns to 1/0
    bool_link_col = [col for col in parameters.bool_col if col in roadway_net.links_df.columns]
    bool_node_col = [col for col in parameters.bool_col if col in roadway_net.nodes_df.columns]

    link_ff_df, link_max_width_dict = dataframe_to_fixed_width(
        roadway_net.links_df[link_output_variables], bool_link_col
    )

    if drive_only:
        link_ff_df = link_ff_df.loc[link_ff_df["drive_access"] == 1]

    WranglerLogger.info("Writing out link database")

    link_ff_df.to_csv(output_link_txt, sep=";", index=False, header=False)

    # write out header and width correspondence
    WranglerLogger.info("Writing out link header and width ----")
    link_max_width_df = DataFrame(
        list(link_max_width_dict.items()), columns=["header", "width"]
    )
    link_max_width_df.to_csv(output_link_header_width_txt, index=False)

    node_ff_df, node_max_width_dict = dataframe_to_fixed_width(
        roadway_net.nodes_df[node_output_variables], bool_node_col
    )
    WranglerLogger.info("Writing out node database")

    if drive_only:
        node_ff_df = node_ff_df.loc[node_ff_df["drive_node"] == 1]

    node_ff_df.to_csv(output_node_txt, sep=";", index=False, header=False)

    # write out header and width correspondence
    WranglerLogger.info("Writing out node header and width")
    node_max_width_df = DataFrame(
        list(node_max_width_dict.items()), columns=["header", "width"]
    )
    node_max_width_df.to_csv(output_node_header_width_txt, index=False)

    # write out cube script
    s = 'RUN PGM = NETWORK MSG = "Read in network from fixed width file" \n'
    s += 'FILEI LINKI[1] = "{}",'.format(output_link_txt)
    start_pos = 1
    for i in range(len(link_max_width_df)):
        s += " VAR=" + link_max_width_df.header.iloc[i]

        if roadway_net.links_df.dtypes.loc[link_max_width_df.header.iloc[i]] == "O":
            s += "(C" + str(link_max_width_df.width.iloc[i]) + ")"

        s += (
            ", BEG="
            + str(start_pos)
            + ", LEN="
            + str(link_max_width_df.width.iloc[i])
            + ","
        )

        start_pos += link_max_width_df.width.iloc[i] + 1

    s = s[:-1]
    s += "\n"
    s += 'FILEI NODEI[1] = "{}",'.format(output_node_txt)
    start_pos = 1
    for i in range(len(node_max_width_df)):
        s += " VAR=" + node_max_width_df.header.iloc[i]

        if roadway_net.nodes_df.dtypes.loc[node_max_width_df.header.iloc[i]] == "O":
            s += "(C" + str(node_max_width_df.width.iloc[i]) + ")"

        s += (
            ", BEG="
            + str(start_pos)
            + ", LEN="
            + str(node_max_width_df.width.iloc[i])
            + ","
        )

        start_pos += node_max_width_df.width.iloc[i] + 1

    s = s[:-1]
    s += "\n"
    s += 'FILEO NETO = "complete_network.net" \n\n'
    s += "ZONES = {} \n\n".format(zones)
    s += "ROADWAY = LTRIM(TRIM(ROADWAY)) \n"
    s += "NAME = LTRIM(TRIM(NAME)) \n"
    s += "\n \nENDRUN"

    with open(output_cube_network_script, "w") as f:
        f.write(s)

# this should be moved to util
# @staticmethod
def dataframe_to_fixed_width(df, bool_col):
    """
    Convert dataframe to fixed width format, geometry column will not be transformed.

    Args:
        df (pandas DataFrame).
        bool_col (boolean column names)

    Returns:
        pandas dataframe:  dataframe with fixed width for each column.
        dict: dictionary with columns names as keys, column width as values.
    """
    WranglerLogger.info("Starting fixed width conversion")

    # get the max length for each variable column
    max_width_dict = dict(
        [
            (v, df[v].apply(lambda r: len(str(r)) if r != None else 0).max())
            for v in df.columns.values
            if v != "geometry"
        ]
    )

    fw_df = df.drop("geometry", axis=1).copy()
    fw_df[bool_col] = fw_df[bool_col].astype(int)

    for c in fw_df.columns:
        fw_df[c] = fw_df[c].apply(lambda x: str(x))
        fw_df["pad"] = fw_df[c].apply(lambda x: " " * (max_width_dict[c] - len(x)))
        fw_df[c] = fw_df.apply(lambda x: x["pad"] + x[c], axis=1)

    return fw_df, max_width_dict

# @staticmethod
def read_match_result(path):
    """
    Reads the shst geojson match returns.

    Returns shst dataframe.

    Reading lots of same type of file and concatenating them into a single DataFrame.

    Args:
        path (str): File path to SHST match results.

    Returns:
        geodataframe: geopandas geodataframe

    ##todo
    not sure why we need, but should be in utilities not this class
    """
    refId_gdf = DataFrame()
    refid_file = glob.glob(path)
    for i in refid_file:
        new = gpd.read_file(i)
        refId_gdf = pd.concat([refId_gdf, new], ignore_index=True, sort=False)
    return refId_gdf

# @staticmethod
def get_attribute(
    links_df,
    join_key,  # either "shstReferenceId", or "shstGeometryId", tests showed the latter gave better coverage
    source_shst_ref_df,  # source shst refId
    source_gdf,  # source dataframe
    field_name,  # , # targetted attribute from source
):
    """
    Gets attribute from source data using SHST match result.

    Args:
        links_df (dataframe): The network dataframe that new attribute should be written to.
        join_key (str): SHST ID variable name used to join source data with network dataframe.
        source_shst_ref_df (str): File path to source data SHST match result.
        source_gdf (str): File path to source data.
        field_name (str): Name of the attribute to get from source data.

    Returns:
        None
    """
    # join based on shared streets geometry ID
    # pp_link_id is shared streets match return
    # source_ink_id is mrcc
    WranglerLogger.debug(
        "source ShSt rename_variables_for_dbf columns\n{}".format(
            source_shst_ref_df.columns
        )
    )
    WranglerLogger.debug("source gdf columns\n{}".format(source_gdf.columns))
    # end up with OSM network with the MRCC Link ID
    # could also do with route_sys...would that be quicker?
    join_refId_df = pd.merge(
        links_df,
        source_shst_ref_df[[join_key, "pp_link_id", "score"]].rename(
            columns={"pp_link_id": "source_link_id", "score": "source_score"}
        ),
        how="left",
        on=join_key,
    )

    # joined with MRCC dataframe to get route_sys

    join_refId_df = pd.merge(
        join_refId_df,
        source_gdf[["LINK_ID", field_name]].rename(
            columns={"LINK_ID": "source_link_id"}
        ),
        how="left",
        on="source_link_id",
    )

    # drop duplicated records with same field value

    join_refId_df.drop_duplicates(
        subset=["model_link_id", "shstReferenceId", field_name], inplace=True
    )

    # more than one match, take the best score

    join_refId_df.sort_values(
        by=["model_link_id", "source_score"],
        ascending=True,
        na_position="first",
        inplace=True,
    )

    join_refId_df.drop_duplicates(
        subset=["model_link_id"], keep="last", inplace=True
    )

    # self.links_df[field_name] = join_refId_df[field_name]

    return join_refId_df[links_df.columns.tolist() + [field_name, "source_link_id"]]
